from typing import Optional
from logging import Logger

from org.codeforrussia.selector.standardizer.common import Standardizer
from org.codeforrussia.selector.standardizer.schemas.schema_registry_factory import \
    StandardProtocolSchemaRegistryFactory


class ShpilkinDumpStandardizer(Standardizer):
    logger = Logger("ShpilkinDumpStandardizer")

    def __init__(self, schema_registry_factory: StandardProtocolSchemaRegistryFactory):
        super().__init__(schema_registry_factory)

    def convert(self, protocol_data: dict) -> Optional[dict]:
        """
        Accepts single-protocol data in Shpilkin's format.
        First, it recognizes the election attributes by election name.
        :param protocol_data: data
        :return: standardized data
        """
        if protocol_data["loc"]:
            election_name = protocol_data["loc"][0]
            election_level, election_type, election_location_type = self.detect_election_attributes_by_name(election_name)
            recognized_schema = self.schema_registry.search_schema(election_level, election_type, election_location_type)
            protocol_fields = self.get_protocol_fields(recognized_schema)

            election_location = protocol_data["loc"][1:]
            election_commission_level = self.detect_commission_level(election_location)
            election_date = "21.09.2020"  # must be included, but no provided in dump; TODO: extract from filename

            sdata = {
                "election": {
                    "name": election_name,
                    "location": election_location,
                    "commission_level": election_commission_level,
                    "date": election_date
                }
            }
            for d in protocol_data['data']:
                if "line_num" in d and "line_val" in d:
                    line_number = int(d["line_num"])
                    line_value = int(d["line_val"])
                    if line_number > 0 and line_number < len(protocol_fields) + 1:
                        sdata[protocol_fields[line_number - 1]] = line_value

            return sdata, election_level, election_type, election_location_type
        else:
            self.logger.warning("Protocol data is likely malformed. Empty result will be returned for this protocol data.")
            return None



from typing import Dict

from org.codeforrussia.selector.standardizer.common import Standardizer
from org.codeforrussia.selector.standardizer.custom.shpilkin import ShpilkinDumpStandardizer
from org.codeforrussia.selector.standardizer.schemas.schema_registry_factory import \
    StandardProtocolSchemaRegistryFactory

from enum import Enum

class SupportedInputFormat(Enum):
    SHPILKIN_DUMP  = "SHPILKIN"
    # TODO: add other custom formats of scraped election data

    def __str__(self):
        return self.value

class StandardizerRegistryFactory(object):

    def get_standardizer_registry(schema_registry_factory: StandardProtocolSchemaRegistryFactory):
        class StandardizerRegistry:
            """
            Registry of supported input data formats and their respective standardizers.
            """
            def __init__(self, schema_registry_factory: StandardProtocolSchemaRegistryFactory):
                self.standardizers: Dict[SupportedInputFormat, Standardizer] = {
                    SupportedInputFormat.SHPILKIN_DUMP : ShpilkinDumpStandardizer(schema_registry_factory=schema_registry_factory)
                }

            def get_standardizer(self, data_format: SupportedInputFormat) -> Standardizer:
                try:
                    return self.standardizers[data_format]
                except KeyError:
                    raise NotImplementedError(f"This input data format is not supported yet: {data_format}")

        return StandardizerRegistry(schema_registry_factory)

    get_standardizer_registry = staticmethod(get_standardizer_registry)


from abc import ABC, abstractmethod
from typing import Optional, Tuple, List, Dict

from org.codeforrussia.selector.standardizer.election_layers import ElectionLevel, ElectionLocationType, ElectionType
from org.codeforrussia.selector.standardizer.schemas.schema_registry_factory import StandardProtocolSchemaRegistryFactory


class Standardizer(ABC):
    PROTOCOL_FIELD_PATTERN = "Строка"

    def __init__(self, schema_registry_factory: StandardProtocolSchemaRegistryFactory):
        self.schema_registry = schema_registry_factory.get_schema_registry()
        self.commission_levels = [f for f in self.schema_registry.get_common_election_schema()["fields"] if f["name"] == "commission_level"][0]["type"]["symbols"]

    def detect_election_attributes_by_name(self, election_name) -> Tuple[ElectionLevel, ElectionType, ElectionLocationType]:
        # TODO: add more accurate recognition; support all combinations
        if "депутатов государственной думы" in election_name.lower():
            return (ElectionLevel.FEDERAL, ElectionType.REPRESENTATIVE, None)
        else:
            raise NotImplementedError(f"Cannot recognize election attributes by this election name: {election_name}")

    def detect_commission_level(self, election_location: [str]) -> str:
        """
        Given election location, recognizes the commission level according to the levels defined in the common schema
        :param election_location: election location
        :return:
        """
        loc_number = len(election_location)
        if loc_number > 0 and loc_number < len(self.commission_levels):
            return self.commission_levels[loc_number]
        else:
            raise ValueError(f"Election location contains wrong number of elements; must be within [1, {len(self.commission_levels) - 1}] got {loc_number}")

    def get_protocol_fields(self, schema) -> [str]:
        """
        Gets protocol fields for a given schema
        :param schema:
        :return:
        """
        return [f['name'] for f in schema['fields'] if 'doc' in f and f['doc'].startswith(self.PROTOCOL_FIELD_PATTERN)]

    def convert_batch(self, batch: List[Dict]) -> List[Optional[Dict]]:
        def tuple_to_dict(output: Optional[Tuple[Dict, ElectionLevel, ElectionType, ElectionLocationType]]) -> Optional[Dict]:
            if output is not None:
                sdata, election_level, election_type, election_location = output
                return {
                    "sdata": sdata,
                    "election_attrs": {"level": election_level, "type": election_type, "location": election_location}
                        }
            else:
                return None

        return [tuple_to_dict(self.convert(protocol_data)) for protocol_data in batch]



    @abstractmethod
    def convert(self, protocol_data: dict) -> Optional[Tuple[dict, ElectionLevel, ElectionType, ElectionLocationType]]:
        pass




from enum import Enum

class ElectionLevel(Enum):
    FEDERAL = 1
    REGIONAL = 2
    MUNICIPAL = 3


class ElectionType(Enum):
    PERSONAL = 1
    REPRESENTATIVE = 2


class ElectionLocationType(Enum):
    MUNICIPAL_DISTRICT = 1
    CITY_RURAL = 2


import unittest
from org.codeforrussia.selector.standardizer.custom.shpilkin import ShpilkinDumpStandardizer
from org.codeforrussia.selector.standardizer.schemas.schema_registry_factory import StandardProtocolSchemaRegistryFactory
import jsonlines
from pathlib import Path

class TextShpilkinDumpStandardizer(unittest.TestCase):
    test_protocol_data = []

    @classmethod
    def setUpClass(cls):
        with jsonlines.open(Path(__file__).parent / "resources" / "shpilkin" / "state_duma_1_0_example.jsonl") as reader:
            for protocol in reader:
                cls.test_protocol_data.append(protocol)

    def test_on_dump(self):
        standardizer = ShpilkinDumpStandardizer(schema_registry_factory=StandardProtocolSchemaRegistryFactory)
        actual = standardizer.convert_batch(TextShpilkinDumpStandardizer.test_protocol_data)
        self.assertEqual(2, len(actual), "Expected length does not match")
        self.assertEqual("TERRITORY", actual[0]["sdata"]["election"]["commission_level"])
        self.assertEqual(18476, actual[0]["sdata"]["valid"])
        self.assertEqual(1183, actual[1]["sdata"]["canceled"])